(defparameter *height* 20)
(defparameter *width* 60)
(defparameter *humans* nil)
(defparameter *init-member* 30)
(defparameter *life* 100)
(defparameter *place* nil)
(defparameter *turn* 0)

(defstruct human
  x y age sex preg)

(defun start ()
  (loop repeat *init-member*
	do (progn 
	     (let ((p (randput)))
	       (push (make-human :x (car p)
				 :y (cdr p)
				 :age (random (ash *life* -1))
				 :sex (random 2)
				 :preg nil)
		     *humans*)
	       (push p *place*))))
  (main-loop))

(defun randput ()
  (let ((p (cons (random *width*) (random *height*))))
    (if (memberp p *place*)
	(randput)
      p)))

(defun memberp (key lst)
  (if (null lst)
      nil
    (if (equal key (car lst))
	t
      (memberp key (cdr lst)))))

(defun main-loop ()
  (show-map)
  (fresh-line)
  (let ((str (read-line)))
    (cond ((equal str "quit") ())
	  (t (let ((x (parse-integer str :junk-allowed t)))
	       (if x
		   (loop
		     for i
		     below x
		     do (update-world)
		     if (zerop (mod i 1000))
		     do (princ #\.))
		 (update-world))
	       (main-loop))))))

(defun show-map ()
  (format t "~&~dN~&" *turn*)
  (princ "|")
  (loop for x
	below *width*
	do (princ #\*))
  (princ "|")
  (loop for y
	below *height*
	do (progn (fresh-line)
		  (princ "|")
		  (loop for x
			below *width*
			do (if (memberp (cons x y) *place*)
			       (let ((h (search-place (cons x y) *humans*)))
				 (if (< (human-age h) 18)
				     (princ #\k)
				   (if (= (human-sex h) 0)
				       (if (< (human-age h) (ash *life* -1))
					   (princ #\f)
					 (princ #\F))
				     (if (< (human-age h) (ash *life* -1))
					 (princ #\m)
				       (princ #\M)))))
			     (princ #\space)))
		  (princ "|")))
  (fresh-line)
  (princ "|")
  (loop for x
	below *width*
	do (princ #\*))
  (princ "|"))

(defun update-world ()
  (incf *turn*)
  (dolist (h *humans*)
    (incf (human-age h))
    (if (human-preg h)
	(bear (human-x h) (human-y h) h)
      (move h))
    (if (< *life* (human-age h))
	(progn (setf *place* (del-key (cons (human-x h) (human-y h)) *place*))
	       (setf *humans* (del-key h *humans*))))))

(defun bear (x y h)
  (cond
   ((not (memberp (cons (1- x) y) *place*))
    (push (make-human :x (1- x)
		      :y y
		      :age 0
		      :sex (random 2)
		      :preg nil)
	  *humans*)
    (setf (human-preg h) nil))
   ((not (memberp (cons (1+ x) y) *place*))
    (push (make-human :x (1+ x)
		      :y y
		      :age 0
		      :sex (random 2)
		      :preg nil)
	  *humans*)
    (setf (human-preg h) nil))
   ((not (memberp (cons x (1+ y)) *place*))
    (push (make-human :x x
		      :y (1+ y)
		      :age 0
		      :sex (random 2)
		      :preg nil)
	  *humans*)
    (setf (human-preg h) nil))
   ((not (memberp (cons x (1- y)) *place*))
    (push (make-human :x x
		      :y (1- y)
		      :age 0
		      :sex (random 2)
		      :preg nil)
	  *humans*)
    (setf (human-preg h) nil))
   (t (human-preg h) nil)))

(defun search-place (p lst)
  (if (null lst)
      nil
    (if (and (= (car p) (human-x (car lst)))
	     (= (cdr p) (human-y (car lst))))
	(car lst)
      (search-place p (cdr lst)))))

(defun move (h)
  (labels ((mv (h)
	       (let* ((x (1- (random 3)))
		      (y (1- (random 3)))
		      (px (+ (human-x h) x))
		      (py (+ (human-y h) y)))
		 (if (not (and (= x 0)
			       (= y 0)))
		     (if (or (memberp (cons px py) *place*)
			     (< px 0)
			     (< (1- *width*) px)
			     (< py 0)
			     (< (1- *height*) py))
			 (move h)
		       (progn (setf *place* (del-key (cons (human-x h) (human-y h)) *place*))
			      (setf (human-x h) px)
			      (setf (human-y h) py)
			      (push (cons px py) *place*)))))))
	  (mv h))
  (if (< 18 (human-age h))
      (cond
       ((scheck h 1 0)
	(if (= (human-sex h) 0)
	    (setf (human-preg h) t)
	  (setf (human-preg (search-place (cons (1+ (human-x h)) (human-y h)) *humans*)) t)))
       ((scheck h -1 0)
	(if (= (human-sex h) 0)
	    (setf (human-preg h) t)
	  (setf (human-preg (search-place (cons (1- (human-x h)) (human-y h)) *humans*)) t)))
       ((scheck h 0 1)
	(if (= (human-sex h) 0)
	    (setf (human-preg h) t)
	  (setf (human-preg (search-place (cons (human-x h) (1+ (human-y h))) *humans*)) t)))
       ((scheck h 0 -1)
	(if (= (human-sex h) 0)
	    (setf (human-preg h) t)
	  (setf (human-preg (search-place (cons (human-x h) (1- (human-y h))) *humans*)) t))))))

(defun scheck (h x y)
  (if (memberp (cons (+ (human-x h) x) (+ (human-y h) y)) *place*)
      (let ((n (search-place (cons (+ (human-x h) x) (+ (human-y h) y)) *humans*)))
	(if (and (/= (human-sex h) (human-sex n))
		 (< 18 (human-age h))
		 (< 18 (human-age n))
		 (< (human-age h) (ash *life* -1))
		 (< (human-age n) (ash *life* -1)))
	    t
	  nil))
    nil))

(defun del-key (p lst)
  (if (null lst)
      nil
    (if (equal p (car lst))
	(del-key p (cdr lst))
      (cons (car lst) (del-key p (cdr lst))))))

(start)